"""
Shared, named numpy arrays
Hans Buehler 2023
"""

from .logger import Logger
from .version import version
import numpy as np
import gc as gc
from multiprocessing import shared_memory
_log = Logger(__file__)

@version("0.0.1")
class ndsharedarray( object ):
    """
    Wrapper around https://docs.python.org/3/library/multiprocessing.shared_memory.html
    Use sharedarray() to create objects of this type
    """

    serial_number = 0

    def __init__(self, name   : str,
                       shape  : tuple, 
                       create : bool,
                       dtype  = np.float32, 
                       full   = 0. ):
        """
        Create a new shared memory array.
        See sharedarray().
        
        Recall that Python objects are not freed until they are garbage collected.
        Make sure you use gc.collect() whenever you want to make sure to delete
        a particular shared memory.

        Parameters
        ----------
            name    : globally unique name accross all processes.
                      Note that the name used for the shared memory itself will be id'd by dtype and shape of the array to avoid collisions.
                      Use shared_id to obtain this name.
            shape   : numpy shape
            create  : whether to create a new shared memory block or not.
            dtype   : numpy dtype
            full    : if not None: value used to fill newly created array. Not used when create is False
        """
        self._dtype  = dtype if isinstance(dtype, type) else type(dtype)
        size         = self._dtype().itemsize * np.product( shape )
        shape        = tuple(shape)
        self._name   = name
        self._id     = name + str(shape) + "[" + str(self._dtype.__name__) + "]"
        self._serial = ndsharedarray.serial_number   # for debugging purposes
        ndsharedarray.serial_number += 1

        assert size>0, "Cannot have zero size"

        if create:
            self._shared = shared_memory.SharedMemory(name=self._id, create=True, size=size )
            print("Created sizes", self._dtype, size, self._shared.size )
            self._array  = np.ndarray( shape, dtype=self._dtype, buffer = self._shared.buf )
            if not full is None:
                self._array[:] = full
        else:
            self._shared = shared_memory.SharedMemory(name=self._id, create=False, size=size )
            self._array  = np.ndarray( shape, dtype=self._dtype, buffer = self._shared.buf )

    def __del__(self):
        """ Ensure shared memory is released """
        if '_shared' in self.__dict__:
            try:
                self._shared.close()
            except FileNotFoundError:
                pass
            try:
               self._shared.unlink()
            except FileNotFoundError:
                pass
            del self._shared
            
    def __str__(self):
        return "ndsharedarray( " + self._id + ")" + str(self._array) 
    
    # Basic
    # -----
    
    @property
    def array(self):
        """ Return underlying numpy array """
        return self._array
    @property
    def shared_id(self):
        """ Return fully qualified name of shared memory, e.g. including dtype and shape information """
        return self._id
    @property
    def shared_size(self):
        """ Return fully qualified name of shared memory, e.g. including dtype and shape information """
        return self._shared.size
    @property
    def name(self):
        """ Name """
        return self._name
    @property
    def shape(self):
        return self._array.shape
    @property
    def dtype(self):
        return self._dtype
    
    # mimic numpy array
    # -----------------
    
    def __getitem__(self, k, *kargs):
        return self._array.__getitem__(k, *kargs)
    def __setitem__(self, k, v, *kargs):
        return self._array.__setitem__(k,v,*kargs)
#    def __getattr__(self, name):
#        return getattr(self._array,name)
    @property
    def __array_interface__(self):
        return self._array.__array_interface__
    def __array__(self, *kargs, **kwargs):
        return self._array.__array__(*kargs, **kwargs)
        
    # pickling
    # --------

    @staticmethod
    def from_state( state ):
        """ Restore object from disk """
        return ndsharedarray( name=state['name'], shape=state['shape'], create=True, dtype=state['dtype'], full=state['data'] )

    def __reduce__(self):
        """
        Pickling this object explicitly
        See https://docs.python.org/3/library/pickle.html#object.__reduce__
        """
        state = dict( name=self._name, 
                      dtype=self._dtype,
                      shape=self.shape,
                      data=self._array
                    )
        return (self.from_state, (state,) )


@version("0.0.1", dependencies=[ndsharedarray] )
def sharedarray( name   : str,
                 shape  : tuple, 
                 create : bool,
                 dtype  = np.float32, 
                 full   = None,
                 *,
                 raiseOnError : bool = False ):
    """
    Create a new shared memory array.

    Recall that Python objects are not freed until they are garbage collected.
    Make sure you use gc.collect() whenever you want to make sure to delete
    a particular shared memory.

    Parameters
    ----------
        name    : globally unique name accross all processes.
        shape   : numpy shape
        create  : whether to create a new shared memory block or not.
                  If True, and if a block with the given name already exists, this function returns None or raises FileExistsError (if raiseOnError is True)
                    Otherwise it will return a new block with the specified name.
                  If False, and no block with the given name exists, this function returns None  or raises FileNotFoundError (if raiseOnError is True)
                    If the block has different total size, the function will raise an IncompatibleSharedSizeError
                    Otherwise the function will return an array pointing to the shared block
                  If None, then the function first attempts to create a new block with the given name.
                      If this is successful, the function will return (array, True).
                    If creating a new memory block fails, it will attempt to read a block with the given name.
                    and return (array, False) where array is pointing to the shared block.
        dtype   : dtype
        full    : If not None, fill a newly created object with this data.
                  Ignored if a shared object is used.
        raiseOnError : if False, fail by returning None, else throw FileExistsError or FileNotFoundError, respectively

    Returns
    -------
        If 'create' is a boolean: returns ndsharedarray or None
        If 'create' is None: return ndsharedarray, created where 'created' is a boolean indicating whether the array was newly created.

    Raises
    ------
        May raise FileExistsError or FileNotFoundError if raiseOnError is True
    """
    name         = str(name)
    create       = bool(create) if not create is None else None

    assert len(name) > 0, "Must specifiy name"

    if create is None or create:
        try:
            array = ndsharedarray( name=name, shape=shape, create=True, dtype=dtype, full=full )
            if not create is None:
                return array
            else:
                return array, True
        except FileExistsError as e:
            if not create is None:
                if raiseOnError:
                    raise e
                return None

    try:
        array = ndsharedarray( name=name, shape=shape, create=False, dtype=dtype, full=None )
        if not create is None:
            return array
        else:
            return array, False
    except FileNotFoundError as e:
        if not create is None:
            if raiseOnError:
                raise e
            return None
    
    raise Exception("Cannot create or read shared memory block '%s'", name) 
        